#include <Arduino.h>
#include <U8g2lib.h>

#define ROWS 4
#define COLUMNS 8
#define NOT_REVEALED 10
#define BOMB 11
#define UMBRELLA 12
#define CAT 13
#define VRX_PIN  A0
#define VRY_PIN  A1
#define UMBRELLA_BUTTON 3
#define REVEAL_BUTTON 2
#define BUZZER 4


U8G2_ST7920_128X64_1_SW_SPI u8g2(U8G2_R0, /* clock=*/ 13, /* data=*/ 11, /* CS=*/ 10, /* reset=*/ 8);

int mapOfPisaVillage[ROWS][COLUMNS];// 10 = not revealed // 11 = cat // 12 = UMBRELLA
int cats[ROWS][COLUMNS] = {0}; // 0 = no cat // 1 = cat
int fieldsRevealed = 0; // game is won when fieldsRevealed = 32 - numberOfcats
int CursorX = 8;
int CursorY = 13;
int gameStatus = 3;
int revealed = 0;
int umbrellasLeft = 0;
int difficulty = -1;

int revealEmptyFields(int i, int j) {
  if (isOnMap(i, j) && mapOfPisaVillage[i][j] == NOT_REVEALED && cats[i][j] == 0) {
    mapOfPisaVillage[i][j] = getNumberOfCatNeighbours(i, j);
    if (mapOfPisaVillage[i][j] == 0) {
      return 1 + revealEmptyFields(i - 1, j) + revealEmptyFields(i - 1, j + 1) +
      revealEmptyFields(i, j + 1) + revealEmptyFields(i + 1, j + 1) +
      revealEmptyFields(i + 1, j) + revealEmptyFields(i + 1, j - 1) +
      revealEmptyFields(i, j - 1) + revealEmptyFields(i - 1, j - 1);
    } else {
      return 1;
    }
  }
  return 0;
}

bool isOnMap(int i, int j) {
  if (i >= 0 && i < ROWS && j >= 0 && j < COLUMNS)
    return true;
  return false;
}

int getNumberOfCatNeighbours(int i, int j) {
  int nr = 0;
  if (isOnMap(i - 1, j) && cats[i - 1][j] == 1)
    nr++;

  if (isOnMap(i - 1, j - 1) && cats[i - 1][j - 1] == 1)
    nr++;

  if (isOnMap(i - 1, j + 1) && cats[i - 1][j + 1] == 1)
    nr++;

  if (isOnMap(i, j - 1) && cats[i][j - 1] == 1)
    nr++;

  if (isOnMap(i, j + 1) && cats[i][j + 1] == 1)
    nr++;

  if (isOnMap(i + 1, j) && cats[i + 1][j] == 1)
    nr++;

  if (isOnMap(i + 1, j - 1) && cats[i + 1][j - 1] == 1)
    nr++;

  if (isOnMap(i + 1, j + 1) && cats[i + 1][j + 1] == 1)
    nr++;

  return nr;
}

void revealAllLost() {
  for (int i = 0; i < ROWS; i++) {
    for (int j = 0; j < COLUMNS; j++) {
      if (cats[i][j] == 1) {
        mapOfPisaVillage[i][j] = BOMB;
      }
      if (mapOfPisaVillage[i][j] == NOT_REVEALED) {
        int neighbour = getNumberOfCatNeighbours(i, j);
        mapOfPisaVillage[i][j] = neighbour;
      }
    }
  }
}

void revealAllWon() {
  for (int i = 0; i < ROWS; i++) {
    for (int j = 0; j < COLUMNS; j++) {
      if (cats[i][j] == 1) {
        mapOfPisaVillage[i][j] = CAT;
      }
      if (mapOfPisaVillage[i][j] == NOT_REVEALED) {
        int neighbour = getNumberOfCatNeighbours(i, j);
        mapOfPisaVillage[i][j] = neighbour;
      }
    }
  }
}

void setup(void) {
  u8g2.begin();
  Serial.begin(9600); // initialize serial
  pinMode(UMBRELLA_BUTTON, INPUT_PULLUP);
  pinMode(REVEAL_BUTTON, INPUT);
  pinMode(BUZZER, OUTPUT);
}

void loop(void) {
  if (gameStatus == 0) {
    //initialize the map
    for (int i = 0; i < ROWS; i++) {
      for (int j = 0; j < COLUMNS; j++) {
        mapOfPisaVillage[i][j] = NOT_REVEALED;
      }
    }
    //generate the cats randomly
    for (int i = 0; i < ROWS; i++) {
      for (int j = 0; j < COLUMNS; j++) {
        cats[i][j] = 0;
      }
    }
    randomSeed(analogRead(A2));
    for (int i = 0; i < difficulty; i++) {
      int row = random(ROWS);
      int col = random(COLUMNS);
      if (cats[row][col] == 1) {
        i--;
      } else {
        cats[row][col] = 1;
      }
    }
    gameStatus = 1; //change game status to PLAYING
  }
  //REVEAL BUTTON
  int revealState = digitalRead(REVEAL_BUTTON);
  if(revealState == HIGH &&
    mapOfPisaVillage[(CursorY-13)/16][(CursorX-8)/16] == NOT_REVEALED
    && gameStatus == 1) {
      if(cats[(CursorY-13)/16][(CursorX-8)/16] == 1) {
        revealAllLost();
        digitalWrite(BUZZER, HIGH);
        delay(200);
        digitalWrite(BUZZER, LOW);
        delay(200);
        digitalWrite(BUZZER, HIGH);
        delay(200);
        digitalWrite(BUZZER, LOW);
        delay(200);
        digitalWrite(BUZZER, HIGH);
        delay(800);
        gameStatus = -1;
      } else {
        digitalWrite(BUZZER, HIGH); // turn on
        delay(100);
        digitalWrite(BUZZER, LOW);
        int neighbour = getNumberOfCatNeighbours((CursorY-13)/16, (CursorX-8)/16);
        if (neighbour == 0) {
          int nr = revealEmptyFields((CursorY-13)/16, (CursorX-8)/16);
          revealed+=nr;
        } else {
          mapOfPisaVillage[(CursorY-13)/16][(CursorX-8)/16] = neighbour;
          revealed++;
        }
      }
  } else {
    digitalWrite(BUZZER, LOW); // turn off
  }

  //UMBRELLA BUTTON
  int buttonState = digitalRead(UMBRELLA_BUTTON);
  if (buttonState == LOW && 
  (mapOfPisaVillage[(CursorY-13)/16][(CursorX-8)/16] == NOT_REVEALED ||
  mapOfPisaVillage[(CursorY-13)/16][(CursorX-8)/16] == UMBRELLA)) {
    if (mapOfPisaVillage[(CursorY-13)/16][(CursorX-8)/16] == NOT_REVEALED &&
      umbrellasLeft > 0) {
      mapOfPisaVillage[(CursorY-13)/16][(CursorX-8)/16] = UMBRELLA;
      umbrellasLeft--;
      delay(100);
    } else if (mapOfPisaVillage[(CursorY-13)/16][(CursorX-8)/16] == UMBRELLA &&
    umbrellasLeft < difficulty) {
      mapOfPisaVillage[(CursorY-13)/16][(CursorX-8)/16] = NOT_REVEALED;
      umbrellasLeft++;
      delay(100);
    }
  }

  //Conditions to win
  if (revealed == ROWS * COLUMNS - difficulty) {
    revealAllWon();
    gameStatus = 2;
  } else {
    int count = 0;
    for (int i = 0; i < ROWS; i++) {
      for (int j = 0; j < COLUMNS; j++) {
        if (mapOfPisaVillage[i][j] == UMBRELLA && cats[i][j] == 1) {
          count++;
        }
      }
    }
    if (count == difficulty) {
      revealAllWon();
      gameStatus = 2;
    }
  }

  if (gameStatus == 3) {
    Menu();
  } else if (gameStatus == 4) {
    Story();
  } else if (gameStatus == 5) {
    LevelSelect();
  }else if (gameStatus == 1){
    BoardGame();
  } else if (gameStatus == -1) {
    LostGame();
  } else if (gameStatus == 2) {
    WonGame();
  } else if (gameStatus == 6) {
    Tutorial();
  }
}

void BoardGame(){
  int xValue = 0 ;
  int yValue = 0 ;
  u8g2.firstPage();
  do {
    //JOYSTICK
    xValue = analogRead(VRX_PIN);
    yValue = analogRead(VRY_PIN);
    if(xValue > 800 && CursorX < 110) {
      CursorX+=16;
      delay(200);
    }
    if(xValue < 400 && CursorX > 12) {
      CursorX-=16;
      delay(200);
    }
    if(yValue > 800 && CursorY < 60) {
      CursorY+=16;
      delay(200);
    }
    if(yValue < 400 && CursorY > 13) {
      CursorY-=16;
      delay(200);
    }
    
    //start of board indexing
    int x = 0;
    int y = 13;
    char buf[7];
    for (int i = 0; i < ROWS; i++) {
      for (int j = 0; j < COLUMNS; j++) {
        if(mapOfPisaVillage[i][j] == NOT_REVEALED) {
          u8g2.setFont(u8g2_font_unifont_t_symbols);
          u8g2.drawUTF8(x + 16*j, y + 16*i, "▇");
        } else if (mapOfPisaVillage[i][j] == UMBRELLA) {
          u8g2.setFont(u8g2_font_unifont_t_symbols);
          u8g2.drawUTF8(x + 16*j, y + 16*i, "☔");
        } else if (mapOfPisaVillage[i][j] == BOMB) {
          u8g2.setFont(u8g2_font_unifont_t_0_76);
          u8g2.drawUTF8(x + 16*j, y + 16*i, "☹");
        } else if (mapOfPisaVillage[i][j] == CAT) {
          u8g2.setFont(u8g2_font_unifont_t_0_76);
          u8g2.drawUTF8(x + 16*j, y + 16*i, "☺");
        } else if (mapOfPisaVillage[i][j] == 0) {
          u8g2.setFont(u8g2_font_unifont_t_0_76);
          u8g2.drawUTF8(x + 16*j, y + 16*i, " ");
        } else {
          u8g2.setFont(u8g2_font_unifont_t_symbols);
          u8g2.drawUTF8(x + 16*j, y + 16*i, itoa(mapOfPisaVillage[i][j], buf, 10));
        }
      }
    }
    if (gameStatus == 1) {
      u8g2.drawUTF8(CursorX, CursorY, "<");
    }

  } while ( u8g2.nextPage() );
}

void Tutorial() {
  u8g2.firstPage();
  do {
    int select = digitalRead(REVEAL_BUTTON);
    if(select == HIGH) {
      gameStatus = 3;
      delay(100);
    }
    u8g2.setFont(u8g2_font_squeezed_r6_tr);
    u8g2.drawStr(36, 14, "Umbrellas = Flags");
    u8g2.drawStr(43, 24, "Cats = Bombs");
    u8g2.drawStr(18, 34, "Green button = Reveal field");
    u8g2.drawStr(22, 44, "Joystick button = Put flag");
    u8g2.drawStr(50, 58, ">  Back");
  } while (u8g2.nextPage());
}

void Story() {
  u8g2.firstPage();
  do {
    int select = digitalRead(REVEAL_BUTTON);
    if(select == HIGH) {
      gameStatus = 3;
      delay(100);
    }
    u8g2.setFont(u8g2_font_squeezed_r6_tr);
    u8g2.drawStr(0, 14, "Pisa Village is beeing attacked by a big");
    u8g2.drawStr(10, 24, "storm. Protect the cats using the");
    u8g2.drawStr(11, 34, "umbrellas. But be careful, if you");
    u8g2.drawStr(4, 44, "accidentally discover a cat, you lose.");
    u8g2.drawStr(50, 58, ">  Back");
  } while (u8g2.nextPage());
}
int menuCursor = 33;
int nr_butt = 0;
int val = 0;
void Menu() {
  u8g2.firstPage();
  do {
    //joystick
    val = analogRead(VRY_PIN);
    if(val < 400 && nr_butt > 0) {
      menuCursor-=10;
      nr_butt--;
      delay(200);
    }
    if(val > 800 && nr_butt < 2) {
      menuCursor+=10;
      nr_butt++;
      delay(200);
    }
    //select button
    int select = digitalRead(REVEAL_BUTTON);
    if(select == HIGH) {
      if (nr_butt == 0) {
        gameStatus = 5; //select level difficulty
      } else if (nr_butt == 1) {
        gameStatus = 4; //story
      } else if (nr_butt == 2) {
        gameStatus = 6; //tutorial
      }
      delay(200);
    }
    u8g2.setFont(u8g2_font_unifont_t_symbols);
    u8g2.drawUTF8(22,15, "Pisasweeper");
    u8g2.setFont(u8g2_font_squeezed_r6_tr);
    u8g2.drawStr(59, 33, "Play");
    u8g2.drawStr(57, 43, "Story");
    u8g2.drawStr(53, 53, "Tutorial");
    u8g2.drawStr(48, menuCursor, ">");
  } while (u8g2.nextPage());
}
int nr_butt2 = 0;
int menuCursor2 = 27;
void LevelSelect() {
  u8g2.firstPage();
  do {
    //joystick
    val = analogRead(VRY_PIN);
    if(val < 400 && nr_butt2 > 0) {
      menuCursor2-=10;
      nr_butt2--;
      delay(200);
    }
    if(val > 800 && nr_butt2 < 3) {
      menuCursor2+=10;
      nr_butt2++;
      delay(200);
    }
    //select button
    int select = digitalRead(REVEAL_BUTTON);
    if(select == HIGH) {
      if (nr_butt2 == 0) {
        gameStatus = 0;
        difficulty = 4; //Easy
        umbrellasLeft = difficulty;
      } else if (nr_butt2 == 1) {
        gameStatus = 0;
        difficulty = 5; //Medium
        umbrellasLeft = difficulty;
      } else if (nr_butt2 == 3) {
        gameStatus = 3;
        nr_butt2 = 0; //Back Button
        menuCursor2 = 25; //Reset cursor for the next visit
      } else if (nr_butt2 == 2) {
        gameStatus = 0;
        difficulty = 7; //Hard
        umbrellasLeft = difficulty;
      }
      delay(200);
    }
    u8g2.setFont(u8g2_font_squeezed_r6_tr);
    u8g2.drawUTF8(22,14, "Please select the difficulty:");
    u8g2.drawStr(59, 27, "Easy");
    u8g2.drawStr(53, 37, "Medium");
    u8g2.drawStr(58, 47, "Hard");
    u8g2.drawStr(59, 57, "Back");
    u8g2.drawStr(48, menuCursor2, ">");
  } while (u8g2.nextPage());
}

void LostGame() {
  digitalWrite(BUZZER, LOW);
  BoardGame();
  delay(4000);
  u8g2.firstPage();
  do {
    digitalWrite(BUZZER, LOW);
    u8g2.setFont(u8g2_font_squeezed_r6_tr);
    u8g2.drawStr(50, 10, "YOU LOST!");
    u8g2.drawStr(28, 30, "The cats are now wet :(");
    u8g2.drawStr(22, 40, "Press 'RESET' to play again.");
  } while (u8g2.nextPage());

  delay(4000);
}

void WonGame(){
  digitalWrite(BUZZER, LOW);
  BoardGame();
  delay(4000);
  u8g2.firstPage();
  do {
    digitalWrite(BUZZER, LOW);
    u8g2.setFont(u8g2_font_squeezed_r6_tr);
    u8g2.drawStr(50, 10, "YOU WON!");
    u8g2.drawStr(30, 30, "The cats are happy :)");
    u8g2.drawStr(22, 40, "Press 'RESET' to play again.");
  } while (u8g2.nextPage());
  delay(4000);
}


